
//
// LoopManager.ck
//
// This attaches a recorder to the output from the synthesizer.
//
// author: Andrew Smith <andrewrs@cs.washington.edu>
// modified: 14 Mar 2011
//

public class LoopManager extends UGen {
    
    // LiSas for recording
    LiSa layers[48];
    
    // Which track are we recording on?
    0 => int currentTrack;
    
    // Are we recording?
    0 => int recording;
    
    // Are we playing back the recordings?
    0 => int playing;
    
    // Input to loop
    UGen @ in;
    
    // Number of beats per minute
    120 => int bpm;
    
    // Number of measures
    4 => int measures;
    
    // Number of beats per measure
    4 => int beatsPerMeasure;
    
    // Metronome to keep track of the time
    Metronome metronome;
    
    // Length of loop
    dur loopLength;
    
    
    //
    // Setup methods: Determine how long to record and how to set the metronome.
    //
    
    public void setBpm(int bpm) {
        bpm => this.bpm;
        bpm => metronome.setBpm;
        
        setupRecord();
    }
    
    public void setMeasure(int beatsPerMeasure, int measures) {
        beatsPerMeasure => this.beatsPerMeasure;
        measures => this.measures;
        
        setupRecord();
    }
    
    // Sets up the LiSa to record for the specified time
    private void setupRecord() {
        (measures * beatsPerMeasure * (1.0 / bpm) * 60)::second => loopLength;
        
        // Set the end of the recording buffer
        for (0 => int i; i < layers.size(); i++) {
            layers[i] => this;
            loopLength => layers[i].duration;
            loopLength => layers[i].loopEndRec;
            loopLength => layers[i].loopEnd;
        }
        
        sendOSCLoopLength();
    }
    
    // Send an OSC message about the loop length
    private void sendOSCLoopLength() {
        OscSend osc;
        // TODO: This should be swapped to multicast
        osc.setHost( "localhost", 48888 );
        osc.startMsg( "/announce/loopDuration", "f" );
        (loopLength / 1::samp) / (second / samp) => osc.addFloat;
        
        <<< "Sent OSC message about the loop duration", "" >>>;
    }
    
    
    
    //
    // Attach to the ChucK pipeline
    //
    
    public UGen chuck(UGen g) {
        // Connections  
        g @=> in;
        g => this;          // Pass-through sound
        metronome => this;  // Add a metronome
        
        // Setup
        for (0 => int i; i < layers.size(); i++) {
            1 => layers[i].loop;
        }
        
        // Connect to the rest of the pipeline
        return this;
    }
    
    
    
    //
    // Recording methods
    //
    
    // Starts recording the current loop
    public void startRecord() {
        if (!recording) {
            <<< "Starting record on track", currentTrack >>>;
            layers[currentTrack].play(0);
    
            // Hook up the current input to the recorder
            in => layers[currentTrack];
        
            // Wrap around the loop in recording
            1 => layers[currentTrack].loopRec;
        
            // Find the right place in the buffer
            now % loopLength => layers[currentTrack].recPos;
        
            // Kick off the record
            layers[currentTrack].record(1);
            metronome.start();
        
            1 => playing;
            1 => recording;
        }
    }
    
    // Stops recording the current loop
    public void stopRecord() {
        if (recording) {
            <<< "Stopping record on track", currentTrack >>>;
            0 => recording;
        
            metronome.stop();
            layers[currentTrack].record(0);
        
            // Disconnect the input
            in =< layers[currentTrack];
        
            // Start playing where we left off.
            now % loopLength => layers[currentTrack].playPos;
            layers[currentTrack].play(1);
        
            // Choose the next track to record to
            if (currentTrack < layers.size() - 1) {
                1 +=> currentTrack;
            }
        }
    }
    
    // Removes the last recorded track
    public void clearLastRecord() {
        if (!recording) {
            if (currentTrack > 0) {
                1 -=> currentTrack;
            }
            
            <<< "Clearing track", currentTrack >>>;
            layers[currentTrack].rampDown(100::ms);
            layers[currentTrack].clear();
        }
    }
    
    
    
    //
    // Playback methods
    //
    
    // Stops recording and all playing tracks
    public void stopAll() {
        stopRecord();
        
        if (playing) {
            for (0 => int i; i < layers.size(); i++) {
                layers[i].play(0);
            }
            
            0 => playing;
        }
    }
    
    // Play all of the recorded tracks
    public void playAll() {
        if (!playing) {
            <<< "Start playing all tracks, up to", currentTrack >>>;
            
            for (0 => int i; i <= currentTrack; i++) {
                0 => layers[i].playPos;
                layers[i].play(1);
            }
            
            1 => playing;
        }
    }
}